/**
 *  Copyright 2016 Paul Querna
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */

package ff

import (
	"encoding/json"
)

// Number struct
type Number struct {
	Int   json.Number
	Float json.Number
}

// NewNumber creates a new number
func NewNumber(e *Number) {
	e.Int = "1"
	e.Float = "3.14"
}
